import os
import os.path

base_path = os.path.abspath(os.path.dirname(__file__))

def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs

    config = Configuration("_extensions", parent_package, top_path)

    ext = "rTE"
    try:
        from Cython.Build import cythonize

        cythonize(os.path.join(base_path, "rTE.pyx"))
    except ImportError:
        pass

    config.add_extension(
        ext, sources=["rTE.cpp", "_rTE.cpp"], include_dirs=[get_numpy_include_dirs()]
    )

    return config



import numpy as np
from scipy.constants import mu_0

def _rTE_forward(frequencies, lamb, sigma, mu, thicknesses):
    """Compute reflection coefficients for Transverse Electric (TE) mode.

    The first layer is considered to be the top most layer. The last
    layer is considered to have infinite thickness. All physical properties 
    are defined starting from the top most layer. 

    Parameters
    ----------
    frequencies : float, numpy.ndarray
        Frequency (Hz); shape = (n_frequency, ).
    lamb : float, numpy.ndarray
        Spatial wavenumber (1/m), shape = (n_filter, ).
    sigma: complex, numpy.ndarray
        Conductivity (S/m), shape = (n_layer, n_frequency).
    mu: complex, numpy.ndarray
        Magnetic permeability (H/m), shape = (n_layer, n_frequency).
    thicknesses: float, numpy.ndarray
        Thickness (m) of each layer, shape = (n_layer-1,).

    Returns
    -------
    rTE: complex numpy.ndarray
        Reflection coefficients, shape = (n_frequency, n_filter)
    """
    n_layer = len(thicknesses)+1

    omega = 2*np.pi*frequencies
    l2 = lamb**2

    k2 = -1j * omega * mu * sigma
    u = np.sqrt(l2 - k2[:, :, None])
    Y = u/(1j*omega*mu)[:, :, None]
    tanh = np.tanh(u[:-1]*thicknesses[:, None, None])
    Yh = Y[-1]
    for k in range(n_layer-2, -1, -1):
        Yh = Y[k] * (Yh + Y[k]*tanh[k])/(Y[k] + Yh*tanh[k])
    # conductivity and susceptibility in air layer is 0
    Y0 = lamb/(1j*omega*mu_0)[:, None]
    TE = (Y0 - Yh)/(Y0 + Yh)

    return TE

def _rTE_gradient(frequencies, lamb, sigma, mu, thicknesses):
    """Compute reflection coefficients for Transverse Electric (TE) mode.

    The first layer is considered to be the top most layer. The last
    layer is considered to have infinite thickness. All physical properties 
    are defined starting from the top most layer. 

    Parameters
    ----------
    frequencies : float, numpy.ndarray
        Frequency (Hz); shape = (n_frequency, ).
    lamb : float, numpy.ndarray
        Spatial wavenumber (1/m), shape = (n_filter, ).
    sigma: complex, numpy.ndarray
        Conductivity (S/m), shape = (n_layer, n_frequency).
    mu: complex, numpy.ndarray
        Magnetic permeability (H/m), shape = (n_layer, n_frequency).
    thicknesses: float, numpy.ndarray
        Thickness (m) of each layer, shape = (n_layer-1,).

    Returns
    -------
    rTE_dsigma: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. conductivity
        shape = (n_layer, n_frequency, n_filter)
    rTE_dh: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. thicknesses
        shape = (n_layer-1, n_frequency, n_filter)
    rTE_dmu: complex numpy.ndarray
        Reflection coefficients gradient w.r.t. magnetic permeability
        shape = (n_layer, n_frequency, n_filter)
    """
    n_frequency = len(frequencies)
    n_filter = len(lamb)
    n_layer = len(thicknesses)+1

    omega = 2*np.pi*frequencies
    l2 = lamb**2

    k2 = -1j * omega * mu * sigma
    u = np.sqrt(l2 - k2[:, :, None])
    Y = u/(1j*omega*mu)[:, :, None]
    tanh = np.tanh(u[:-1]*thicknesses[:, None, None])
    Yh = np.empty((n_layer, n_frequency, n_filter), dtype=np.complex128)
    Yh[-1] = Y[-1]
    for k in range(n_layer-2, -1, -1):
        Yh[k] = Y[k] * (Yh[k+1] + Y[k]*tanh[k])/(Y[k] + Yh[k+1]*tanh[k])
    Y0 = lamb/(1j*omega*mu_0)[:, None]
    # TE = (Y0 - Yh[0])/(Y0 + Yh[0])

    rTE_dsigma = np.empty((n_layer, n_frequency, n_filter), dtype=np.complex128)
    rTE_dh = np.empty((n_layer-1, n_frequency, n_filter), dtype=np.complex128)
    rTE_dmu = np.empty((n_layer, n_frequency, n_filter), dtype=np.complex128)

    gyh0 = -2.0*Y0/((Y0 + Yh[0])*(Y0 + Yh[0]))
    for k in range(n_layer-1):
        bot = (Y[k] + Yh[k+1]*tanh[k])*(Y[k] + Yh[k+1]*tanh[k])
        gy = gyh0 * tanh[k]*(2.0*tanh[k]*Y[k]*Yh[k+1] + Y[k]*Y[k] + Yh[k+1]*Yh[k+1])/bot
        gtanh = gyh0 * (Y[k]*Y[k]*Y[k] - Y[k]*Yh[k+1]*Yh[k+1])/bot
        gyh0 = gyh0 * -(tanh[k]*tanh[k] - 1.0)*Y[k]*Y[k]/bot

        rTE_dh[k] = gtanh * u[k] * (1.0 - tanh[k]*tanh[k])
        gu = gtanh * thicknesses[k] * (1.0 - tanh[k]*tanh[k])

        gu += gy/(1j * omega * mu[k])[:, None]

        gmu = gy * (-Y[k]/mu[k][:, None])
        gk2 = gu * -0.5 / u[k]

        gmu -= gk2 * (1j * omega * sigma[k])[:, None]
        rTE_dsigma[k] = gk2 * (-1j * omega * mu[k])[:, None]
        rTE_dmu[k] = gmu
    gu = gyh0 / (1j * omega * mu[-1])[:, None]
    gmu = gyh0 * (-Yh[-1]/mu[-1][:, None])
    gk2 = gu * -0.5 / u[-1]
    gmu += gk2 * (-1j * omega * sigma[-1])[:, None]

    rTE_dsigma[-1] = gk2 * (-1j * omega * mu[-1])[:, None]
    rTE_dmu[-1] = gmu

    return rTE_dsigma, rTE_dh, rTE_dmu


try:
    from geoana.kernels._extensions.rTE import rTE_forward, rTE_gradient
except ImportError:
    # Store the above as the kernels
    rTE_forward = _rTE_forward
    rTE_gradient = _rTE_gradient



import numpy as np


def mkvc(x, numDims=1):
    """Creates a vector with the number of dimension specified

    e.g.::

        a = np.array([1, 2, 3])

        mkvc(a, 1).shape
            > (3, )

        mkvc(a, 2).shape
            > (3, 1)

        mkvc(a, 3).shape
            > (3, 1, 1)

    """
    if isinstance(x, np.matrix):
        x = np.array(x)

    if hasattr(x, 'tovec'):
        x = x.tovec()

    assert isinstance(x, np.ndarray), "Vector must be a numpy array"

    if numDims == 1:
        return x.flatten(order='F')
    elif numDims == 2:
        return x.flatten(order='F')[:, np.newaxis]
    elif numDims == 3:
        return x.flatten(order='F')[:, np.newaxis, np.newaxis]


def ndgrid(*args, **kwargs):
    """
    Form tensorial grid for 1, 2, or 3 dimensions.

    Returns as column vectors by default.

    To return as matrix input:

        ndgrid(..., vector=False)

    The inputs can be a list or separate arguments.

    e.g.::

        a = np.array([1, 2, 3])
        b = np.array([1, 2])

        XY = ndgrid(a, b)
            > [[1 1]
               [2 1]
               [3 1]
               [1 2]
               [2 2]
               [3 2]]

        X, Y = ndgrid(a, b, vector=False)
            > X = [[1 1]
                   [2 2]
                   [3 3]]
            > Y = [[1 2]
                   [1 2]
                   [1 2]]

    """

    # Read the keyword arguments, and only accept a vector=True/False
    vector = kwargs.pop('vector', True)
    assert isinstance(vector, bool), "'vector' keyword must be a bool"
    assert len(kwargs) == 0, "Only 'vector' keyword accepted"

    # you can either pass a list [x1, x2, x3] or each seperately
    if type(args[0]) == list:
        xin = args[0]
    else:
        xin = args

    # Each vector needs to be a numpy array
    assert np.all(
        [isinstance(x, np.ndarray) for x in xin]
    ), "All vectors must be numpy arrays."

    if len(xin) == 1:
        return xin[0]
    elif len(xin) == 2:
        XY = np.broadcast_arrays(mkvc(xin[1], 1), mkvc(xin[0], 2))
        if vector:
            X2, X1 = [mkvc(x) for x in XY]
            return np.c_[X1, X2]
        else:
            return XY[1], XY[0]
    elif len(xin) == 3:
        XYZ = np.broadcast_arrays(
            mkvc(xin[2], 1), mkvc(xin[1], 2), mkvc(xin[0], 3)
        )
        if vector:
            X3, X2, X1 = [mkvc(x) for x in XYZ]
            return np.c_[X1, X2, X3]
        else:
            return XYZ[2], XYZ[1], XYZ[0]



print("import: 'geoana'")
import geoana

print("import: 'geoana.earthquake'")
import geoana.earthquake

print("import: 'geoana.em'")
import geoana.em

print("import: 'geoana.kernels._extensions'")
import geoana.kernels._extensions




import numpy as np

def layered_potential_anisotropy_tilde(wx, wy, sxx, sxy, sxz, syy, syz, szz, thicknesses):
    """
    Caλculates phi_tilde for unitary source at the surface.

    wx, wy
    """
    i0 = np.where((wx == 0) & (wy == 0))[0]
    wx[i0] = 1.0
    wy[i0] = 1.0
    N = len(sxx)

    a = szz[:, None]
    b = sxz[:, None]*wx+syz[-1][:, None]*wy
    c = sxx[:, None]*wx*wx+2*sxy[:, None]*wx*wy+syy[:, None]*wy*wy

    aλ = np.sqrt(a*c-b*b)
    tanh = np.tanh(aλ[:-1] * thicknesses[:, None] / a)
    Q = 1/aλ[-1]
    for i in range(N-2, -1, -1):
        aλQ = aλ[i]*Q
        Q = (aλQ+tanh[i])/(aλ*(aλQ*tanh[i]+1))
    Q[i0] = 0.0
    return Q/(2*np.pi)

def grad_layered_potential_anisotropy_tilde(wx, wy, sxx, sxy, sxz, syy, syz, szz, thicknesses):
    """
    Caλculates phi_tilde for unitary source at the surface.

    wx, wy
    """
    i0 = np.where((wx == 0) & (wy==0))[0]
    wx[i0] = 1.0
    wy[i0] = 1.0
    N = len(sxx)

    a = szz[:, None]
    b = sxz[:, None]*wx+syz[-1][:, None]*wy
    c = sxx[:, None]*wx*wx+2*sxy[:, None]*wx*wy+syy[:, None]*wy*wy

    aλ = np.sqrt(a*c-b*b)
    tanh = np.tanh(aλ[:-1] * thicknesses[:, None] / a[:-1])
    Q = np.empty((N, len(wx)))
    Q[-1] = 1/aλ[-1]
    for i in range(N-2, -1, -1):
        aλQ = aλ[i]*Q[i+1]
        Q[i] = (aλQ+tanh[i])/(aλ*(aλQ*tanh[i]+1))
    Q[:, i0] = 0.0
    Q /= (2*np.pi)

    # a = szz[-1]
    # b = sxz[-1]*wx+syz[-1]*wy
    # c = sxx[-1]*wx*wx+2*sxy[-1]*wx*wy+syy[-1]*wy*wy
    # aλ = np.sqrt(a*c-b*b)
    # Q = 1/(aλ)
    # for i in range(N-2, -1, -1):
    #     a = szz[i]
    #     b = sxz[i]*wx+syz[i]*wy
    #     c = sxx[i]*wx*wx+2*sxy[i]*wx*wy+syy[i]*wy*wy
    #     aλ = np.sqrt(a*c-b*b)
    #     λ = aλ/a
    #     tanh = np.tanh(thicknesses[i]*λ)
    #     aλQ = aλ*Q
    #     Q = (aλQ+tanh)/(aλ*(aλQ*tanh+1))
    # Q /= (2*np.pi)

    g_a = np.empty((N, len(wx)))
    g_b = np.empty((N, len(wx)))
    g_c = np.empty((N, len(wx)))
    Q_dh = np.empty((N-1, len(wx)))
    g_Q = 1
    for i in range(N-1):
        aλQ = aλ[i]*Q[i+1]
        g_aλQ = (1 - tanh[i]**2)/(aλ*(aλQ*tanh[i]+1)**2)*g_Q
        g_aλ = -Q[i]/aλ*g_Q
        g_tanh = (1 - aλQ**2)/(aλ*(aλQ*tanh[i]+1)**2)*g_Q
        g_aλ += Q[i+1]*g_aλQ
        g_Q = aλ[i]*g_aλQ
        g_λ = (1.0 - tanh[i]**2) * thicknesses[i] * g_tanh
        Q_dh[i] = (1.0 - tanh[i]**2) * aλ[i] / a[i] * g_tanh
        g_a[i] = -aλ[i]/(a[i]*a[i])*g_λ
        g_aλ += g_λ/a
        g_a[i] += 0.5*c[i] / aλ[i] * g_aλ
        g_b[i] = -b[i] / aλ[i] * g_aλ
        g_c[i] = 0.5 * a[i] / aλ[i] * g_aλ
    g_aλ = -2 / aλ[i] * g_Q
    g_a[-1] += 0.5 * c[-1] / aλ[-1] * g_aλ
    g_b[-1] = -b[-1] / aλ[-1] * g_aλ
    g_c[-1] = 0.5 * a[-1] / aλ[-1] * g_aλ

    Q_sxx = wx*wx * g_c
    Q_sxy = 2*wx*wy * g_c
    Q_syy = wy * wy * g_c
    Q_sxz = wx * g_b
    Q_syz = wy * g_b
    Q_szz = g_a
    return Q_sxx, Q_sxy, Q_sxz, Q_syy, Q_syz, Q_szz



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import properties
from scipy.special import ellipk, ellipe

from .base import BaseEM, BaseDipole, BaseMagneticDipole
from .. import spatial

__all__ = [
    "MagneticDipoleWholeSpace", "CircularLoopWholeSpace",
    "MagneticPoleWholeSpace"
]


class MagneticDipoleWholeSpace(BaseMagneticDipole, BaseEM):
    """
    Static magnetic dipole in a wholespace.
    """

    def vector_potential(self, xyz, coordinates="cartesian"):
        """Vector potential of a static magnetic dipole. See Griffiths, 1999
        equation 5.83

        .. math::

            \\vec{A}(\\vec{r}) = \\frac{\mu_0}{4\pi}
            \\frac{\\vec{m}\\times\\vec{r}}{r^3}

        **Required**

        :param numpy.ndarray xyz: Location at which we calculate the vector
                                potential

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic vector potential at each observation location

        """
        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        # orientation of the dipole
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        dxyz = self.vector_distance(xyz)
        r = spatial.repeat_scalar(self.distance(xyz))
        m = self.moment * np.atleast_2d(self.orientation).repeat(n_obs, axis=0)

        m_cross_r = np.cross(m, dxyz)
        a = (self.mu / (4 * np.pi)) * m_cross_r / (r**3)

        if coordinates.lower() == "cylindrical":
            a = spatial.cartesian_2_cylindrical(xyz, a)

        return a

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        """Magnetic flux (:math:`\\vec{b}`) of a static magnetic dipole

        **Required**

        :param numpy.ndarray xyz: Location of the receivers(s)

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic flux at each observation location
        """

        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        r = self.vector_distance(xyz)
        dxyz = spatial.repeat_scalar(self.distance(xyz))
        m_vec = (
            self.moment * np.atleast_2d(self.orientation).repeat(n_obs, axis=0)
        )

        m_dot_r = (m_vec * r).sum(axis=1)

        # Repeat the scalars
        m_dot_r = np.atleast_2d(m_dot_r).T.repeat(3, axis=1)
        # dxyz = np.atleast_2d(dxyz).T.repeat(3, axis=1)

        b = (self.mu / (4 * np.pi)) * (
            (3.0 * r * m_dot_r / (dxyz ** 5)) -
            m_vec / (dxyz ** 3)
        )

        if coordinates.lower() == "cylindrical":
            b = spatial.cartesian_2_cylindrical(xyz, b)

        return b

    def magnetic_field(self, xyz, coordinates="cartesian"):
        """Magnetic field (:math:`\\vec{h}`) of a static magnetic dipole

        **Required**

        :param numpy.ndarray xyz: Location of the receivers(s)

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic field at each observation location

        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu


class MagneticPoleWholeSpace(BaseMagneticDipole, BaseEM):
    """
    Static magnetic pole in a wholespace.
    """

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        """Magnetic flux (:math:`\\vec{b}`) of a static magnetic dipole

        **Required**

        :param numpy.ndarray xyz: Location of the receivers(s)

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic flux at each observation location
        """

        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        r = self.vector_distance(xyz)
        dxyz = spatial.repeat_scalar(self.distance(xyz))

        b = self.moment * self.mu / (4 * np.pi * (dxyz ** 3)) * r

        if coordinates.lower() == "cylindrical":
            b = spatial.cartesian_2_cylindrical(xyz, b)

        return b

    def magnetic_field(self, xyz, coordinates="cartesian"):
        """Magnetic field (:math:`\\vec{h}`) of a static magnetic dipole

        **Required**

        :param numpy.ndarray xyz: Location of the receivers(s)

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic field at each observation location

        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu


class CircularLoopWholeSpace(BaseDipole, BaseEM):

    """
    Static magnetic field from a circular loop in a wholespace.
    """

    current = properties.Float(
        "Electric current through the loop (A)", default=1.
    )

    radius = properties.Float(
        "radius of the loop (m)", default=1., min=0.
    )

    def vector_potential(self, xyz, coordinates="cartesian"):
        """Vector potential due to the a steady-state current through a
        circular loop. We solve in cylindrical coordinates

        .. math::

            A_\\theta(\\rho, z) = \\frac{\mu_0 I}{\pi k}
            \sqrt{R / \\rho^2}[(1 - k^2/2) * K(k^2) - K(k^2)]

        where

        .. math::

            k^2 = \\frac{4 R \\rho}{(R + \\rho)^2 + z^2}

        and

        - :math:`\\rho = \sqrt{x^2 + y^2}` is the horizontal distance to the test point
        - :math:`r` is the distance to a test point
        - :math:`I` is the current through the loop
        - :math:`R` is the radius of the loop
        - :math:`E(k^2)` and :math:`K(k^2)` are the complete elliptic integrals


        **Required**

        :param numpy.ndarray xyz: Location where we calculate the vector
                                potential

        **Optional**

        :param str coordinates: coordinate system that the xyz is provided
                                in and that the solution will be returned
                                in (cartesian or cylindrical).
                                Default: `"cartesian"`

        **Returns**

        :rtype: numpy.ndarray
        :return: The magnetic vector potential at each observation location

        """
        eps = 1e-10
        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        # convert coordinates if not cartesian
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        xyz = spatial.rotate_points_from_normals(
            xyz, np.array(self.orientation),  # work around for a properties issue
            np.r_[0., 0., 1.], x0=np.array(self.location)
        )

        n_obs = xyz.shape[0]
        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)

        rho = np.sqrt((dxyz[:, :2]**2).sum(1))

        k2 = (4 * self.radius * rho) / ((self.radius + rho)**2 +dxyz[:, 2]**2)
        k2[k2 > 1.] = 1.  # if there are any rounding errors

        E = ellipe(k2)
        K = ellipk(k2)

        # singular if rho = 0, k2 = 1
        ind = (rho > eps) & (k2 < 1)

        Atheta = np.zeros_like(r)
        Atheta[ind] = (
            (self.mu * self.current) / (np.pi * np.sqrt(k2[ind])) *
            np.sqrt(self.radius / rho[ind]) *
            ((1. - k2[ind] / 2.)*K[ind] - E[ind])
        )

        # assume that the z-axis aligns with the polar axis
        A = np.zeros_like(xyz)
        A[ind, 0] = Atheta[ind] * (-dxyz[ind, 1] / rho[ind])
        A[ind, 1] = Atheta[ind] * (dxyz[ind, 0] / rho[ind])

        # rotate the points to aligned with the normal to the source
        A = spatial.rotate_points_from_normals(
            A, np.r_[0., 0., 1.], np.array(self.orientation),
            x0=np.array(self.location)
        )

        if coordinates.lower() == "cylindrical":
            A = spatial.cartesian_2_cylindrical(xyz, A)

        return A

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        """Calculates the magnetic flux density (B) due to a circular current loop

        Parameters
        ----------
        xyz : np.ndarray
            locations to evaluate the function at shape (3, ) or (*, 3)
        coordinates : {cartesian, cylindrical}
            which coordinate system the input and output points are defined in.

        Returns
        -------
        B_field : np.ndarray
            Magnetic Flux Density vector at the given points, shape (*, 3)
        """
        xyz = np.atleast_2d(xyz)
        # convert coordinates if not cartesian
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)
        elif coordinates.lower() != "cartesian":
            raise TypeError(
                f"coordinates must be 'cartesian' or 'cylindrical', the coordinate "
                f"system you provided, {coordinates}, is not yet supported."
            )

        xyz = spatial.rotate_points_from_normals(
            xyz, np.array(self.orientation),  # work around for a properties issue
            np.r_[0., 0., 1.], x0=np.array(self.location)
        )
        # rotate all the points such that the orientation is directly vertical

        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)

        rho = np.linalg.norm(dxyz[:, :2], axis=-1)

        B = np.zeros((len(rho), 3))

        # for On axis points
        inds_axial = rho==0.0

        B[inds_axial, -1] = self.mu * self.current * self.radius**2 / (
            2 * (self.radius**2 + dxyz[inds_axial, 2]**2)**(1.5)
        )

        # Off axis
        alpha = rho[~inds_axial]/self.radius
        beta = dxyz[~inds_axial, 2]/self.radius
        gamma = dxyz[~inds_axial, 2]/rho[~inds_axial]

        Q = ((1+alpha)**2 + beta**2)
        k2 =  4 * alpha/Q

        # axial part:
        B[~inds_axial, -1] = self.mu * self.current / (2 * self.radius * np.pi * np.sqrt(Q)) * (
            ellipe(k2)*(1 - alpha**2 - beta**2)/(Q  - 4 * alpha) + ellipk(k2)
        )

        # radial part:
        B_rad = self.mu * self.current * gamma / (2 * self.radius * np.pi * np.sqrt(Q)) * (
            ellipe(k2)*(1 + alpha**2 + beta**2)/(Q  - 4 * alpha) - ellipk(k2)
        )

        # convert radial component to x and y..
        B[~inds_axial, 0] = B_rad * (dxyz[~inds_axial, 0]/rho[~inds_axial])
        B[~inds_axial, 1] = B_rad * (dxyz[~inds_axial, 1]/rho[~inds_axial])

        # rotate the vectors to be aligned with the normal to the source
        B = spatial.rotate_points_from_normals(
           B, np.r_[0., 0., 1.], np.array(self.orientation),
        )

        if coordinates.lower() == "cylindrical":
            B = spatial.cartesian_2_cylindrical(xyz, B)

        return B

    def magnetic_field(self, xyz, coordinates="cartesian"):
        """Calculates the magnetic field (H) due to a circular current loop

        Parameters
        ----------
        xyz : np.ndarray
            locations to evaluate the function at shape (3, ) or (*, 3)
        coordinates : {cartesian, cylindrical}
            which coordinate system the input and output points are defined in.

        Returns
        -------
        H_field : np.ndarray
            Magnetic Field vector at the given points, shape (*, 3)
        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu



import numpy as np
import properties
from scipy.constants import mu_0, epsilon_0

from .. import spatial


###############################################################################
#                                                                             #
#                              Base Classes                                   #
#                                                                             #
###############################################################################


class BaseEM(properties.HasProperties):
    """
    Base class for electromanetics. Contains physical properties that are
    relevant to all problems that use Maxwell's equations
    """

    mu = properties.Float(
        "Magnetic permeability (H/m)",
        default=mu_0,
        min=0.0
    )

    sigma = properties.Float(
        "Electrical conductivity (S/m)",
        default=1.0,
        min=0.0
    )

    epsilon = properties.Float(
        "Permitivity value (F/m)",
        default=epsilon_0,
        min=0.0
    )


class BaseDipole(BaseEM):
    """
    Base class for dipoles.
    """

    orientation = properties.Vector3(
        "orientation of dipole",
        default="X",
        length=1.0
    )

    location = properties.Vector3(
        "location of the electric dipole source",
        default="ZERO"
    )

    def vector_distance(self, xyz):
        """
        Vector distance from the dipole location
        :param numpy.ndarray xyz: grid
        """
        return spatial.vector_distance(xyz, np.array(self.location))

    def distance(self, xyz):
        """
        Distance from the dipole location
        """
        return spatial.distance(xyz, np.array(self.location))

    def dot_orientation(self, xyz):
        """
        Take the dot product between a grid and the orientation of the dipole
        """
        return spatial.vector_dot(xyz, np.array(self.orientation))

    def cross_orientation(self, xyz):
        """
        Take the cross product between a grid and the orientation of the dipole
        """
        orientation = np.kron(
            np.atleast_2d(
                np.array(self.orientation)
            ), np.ones((xyz.shape[0], 1))
        )
        return np.cross(xyz, orientation)


class BaseElectricDipole(BaseDipole):
    """
    Base class for electric current dipoles
    """

    length = properties.Float(
        "length of the dipole (m)",
        default=1.0,
        min=0.0
    )

    current = properties.Float(
        "magnitude of the injected current (A)",
        default=1.0,
        min=0.0
    )


class BaseMagneticDipole(BaseDipole):
    """
    Base class for magnetic dipoles
    """

    moment = properties.Float(
        "moment of the dipole (Am^2)",
        default=1.0,
        min=0.0
    )



import properties
import numpy as np
from geoana.em.base import BaseMagneticDipole
from geoana.em.fdem.base import BaseFDEM, sigma_hat
from scipy.constants import mu_0, epsilon_0
from empymod.utils import check_hankel
from empymod.transform import get_dlf_points
from geoana.kernels.tranverse_electric_reflections import rTE_forward


class MagneticDipoleLayeredHalfSpace(BaseMagneticDipole, BaseFDEM):

    thickness = properties.Array(
        "Layer thicknesses (m) starting from the top-most layer. The bottom layer is assumed to be infinite.",
        shape=('*', ),
        dtype=float
    )

    sigma = properties.Array(
        "Electrical conductivity (S/m), defined starting from the top most layer",
        shape=('*', ),
        dtype=complex,
        coerce=True
    )

    frequency = properties.Array(
        "Source frequency (Hz)",
        shape=('*', ),
        dtype=float
    )

    mu = properties.Array(
        "Magnetic permeability (H/m), defined starting from the top most layer",
        shape=('*', ),
        dtype=complex,
        default=np.array([mu_0], dtype=np.complex128)
    )

    epsilon = properties.Array(
        "Permitivity value (F/m), defined starting from the top most layer",
        shape=('*', ),
        dtype=float,
        default=np.array([epsilon_0], dtype=np.float64)
    )

    def _get_valid_properties(self):
        thick = self.thickness
        n_layer = len(thick)+1
        sigma = self.sigma
        epsilon = self.epsilon
        mu = self.mu
        if n_layer != 1:
            sigma = self.sigma
            if len(sigma) == 1:
                sigma = np.ones(n_layer)*sigma
            epsilon = self.epsilon
            if len(epsilon) == 1:
                epsilon = np.ones(n_layer)*epsilon
            mu = self.mu
            if len(mu) == 1:
                mu = np.ones(n_layer)*mu
        return thick, sigma, epsilon, mu

    @property
    def sigma_hat(self):
        _, sigma, epsilon, _ = self._get_valid_properties()
        return sigma_hat(
            self.frequency[:, None], sigma, epsilon,
            quasistatic=self.quasistatic
        ).T

    @property
    def wavenumber(self):
        raise NotImplementedError()

    @property
    def skin_depth(self):
        raise NotImplementedError()

    def magnetic_field(self, xyz, field="secondary"):
        """
        Magnetic field due to a magnetic dipole in a layered halfspace at a specific height z

        Parameters
        ----------
        xyz : numpy.ndarray
            receiver locations of shape (n_locations, 3).
            The z component cannot be below the surface (z=0.0).
        field : ("secondary", "total")
            Flag for the type of field to return.
        """

        if np.any(xyz[:, 2] < 0.0):
            raise ValueError("Cannot compute fields below the surface")
        h = self.location[2]
        dxyz = xyz - self.location
        offsets = np.linalg.norm(dxyz[:, :-1], axis=-1)

        # Comput transform operations
        # -1 gives lagged convolution in dlf
        ht, htarg = check_hankel('dlf', {'dlf': 'key_101_2009', 'pts_per_dec': 0}, 1)
        fhtfilt = htarg['dlf']
        pts_per_dec = htarg['pts_per_dec']

        f = self.frequency
        n_frequency = len(f)

        lambd, int_points = get_dlf_points(fhtfilt, offsets, pts_per_dec)

        thick = self.thickness
        n_layer = len(thick) + 1

        thick, sigma, epsilon, mu = self._get_valid_properties()
        sigh = sigma_hat(
            self.frequency[:, None], sigma, epsilon,
            quasistatic=self.quasistatic
        ).T  # this gets sigh with proper shape (n_layer x n_freq) and fortran ordering.
        mu = np.tile(mu, (n_frequency, 1)).T  # shape(n_layer x n_freq)

        rTE = rTE_forward(f, lambd.reshape(-1), sigh, mu, thick)
        rTE = rTE.reshape((n_frequency, *lambd.shape))

        # secondary is height of receiver plus height of source
        rTE *= np.exp(-lambd*(xyz[:, -1] + h)[:, None])
        # works for variable xyz because each point has it's own lambdas

        src_x, src_y, src_z = self.orientation
        C0x = C0y = C0z = 0.0
        C1x = C1y = C1z = 0.0
        if src_x != 0.0:
            C0x += src_x*(dxyz[:, 0]**2/offsets**2)[:, None]*lambd**2
            C1x += src_x*(1/offsets - 2*dxyz[:, 0]**2/offsets**3)[:, None]*lambd
            C0y += src_x*(dxyz[:, 0]*dxyz[:, 1]/offsets**2)[:, None]*lambd**2
            C1y -= src_x*(2*dxyz[:, 0]*dxyz[:, 1]/offsets**3)[:, None]*lambd
            # C0z += 0.0
            C1z -= (src_x*dxyz[:, 0]/offsets)[:, None]*lambd**2

        if src_y != 0.0:
            C0x += src_y*(dxyz[:, 0]*dxyz[:, 1]/offsets**2)[:, None]*lambd**2
            C1x -= src_y*(2*dxyz[:, 0]*dxyz[:, 1]/offsets**3)[:, None]*lambd
            C0y += src_y*(dxyz[:, 1]**2/offsets**2)[:, None]*lambd**2
            C1y += src_y*(1/offsets - 2*dxyz[:, 1]**2/offsets**3)[:, None]*lambd
            # C0z += 0.0
            C1z -= (src_y*dxyz[:, 1]/offsets)[:, None]*lambd**2

        if src_z != 0.0:
            # C0x += 0.0
            C1x += (src_z*dxyz[:, 0]/offsets)[:, None]*lambd**2
            # C0y += 0.0
            C1y += (src_z*dxyz[:, 1]/offsets)[:, None]*lambd**2
            C0z += src_z*lambd**2
            # C1z += 0.0

        # Do the hankel transform on each component
        em_x = ((C0x*rTE)@fhtfilt.j0 + (C1x*rTE)@fhtfilt.j1)/offsets
        em_y = ((C0y*rTE)@fhtfilt.j0 + (C1y*rTE)@fhtfilt.j1)/offsets
        em_z = ((C0z*rTE)@fhtfilt.j0 + (C1z*rTE)@fhtfilt.j1)/offsets

        if field == "total":
            # add in the primary field
            r = np.linalg.norm(dxyz, axis=-1)
            mdotr = src_x*dxyz[:, 0] + src_y*dxyz[:, 1] + src_z*dxyz[:, 2]

            em_x += 3*dxyz[:, 0]*mdotr/r**5 - src_x/r**3
            em_y += 3*dxyz[:, 1]*mdotr/r**5 - src_y/r**3
            em_z += 3*dxyz[:, 2]*mdotr/r**5 - src_z/r**3

        return self.moment/(4*np.pi)*np.stack((em_x, em_y, em_z), axis=-1)



from geoana.em.base import BaseEM
import numpy as np
from scipy.constants import mu_0
import properties


###############################################################################
#                                                                             #
#                           Utility Functions                                 #
#                                                                             #
###############################################################################

def peak_time(z, sigma, mu=mu_0):
    """
    `Peak time <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peaktime.html>`_:
    Time at which the maximum signal amplitude is observed at a particular
    location for a transient plane wave through a homogeneous medium.


    **Required**

    :param float z: distance from source (m)
    :param float sigma: electrical conductivity (S/m)

    **Optional**

    :param float mu: magnetic permeability (H/m). Default: :math:`\mu_0 = 4\pi \\times 10^{-7}` H/m

    """
    return (mu * sigma * z**2)/6.


def diffusion_distance(time, sigma, mu=mu_0):
    """
    `Diffusion distance <https://em.geosci.xyz/content/maxwell1_fundamentals/transient_planewaves_homogeneous/peakdistance.html>`_:
    Distance at which the signal amplitude is largest for a given time after
    shut off. Also referred to as the peak distance
    """
    return np.sqrt(2*time/(mu*sigma))


def theta(time, sigma, mu=mu_0):
    """
    Analog to wavenumber in the frequency domain. See Ward and Hohmann, 1988
    pages 174-175
    """
    return np.sqrt(mu*sigma/(4.*time))


class BaseTDEM(BaseEM):

    time = properties.Float(
        "time after shut-off at which we are evaluating the fields (s)",
        required=True,
        default=1e-4
    )

    def peak_time(self, z):
        return peak_time(z, self.sigma, self.mu)

    @property
    def diffusion_distance(self):
        return diffusion_distance(self.time, self.sigma, self.mu)

    @property
    def theta(self):
        return theta(self.time, self.sigma, self.mu)



import numpy as np
from scipy.constants import mu_0


def vertical_magnetic_field_horizontal_loop(
    frequencies, sigma=1.0, mu=mu_0, radius=1.0, current=1.0, turns=1, secondary=True
):
    """Vertical magnetic field at the center of a horizontal loop for each frequency

    Simple function to calculate the vertical magnetic field due to a harmonic
    horizontal loop. The anlytic form is only available at the center of the loop.

    Parameters
    ----------
    frequencies : float, or numpy.ndarray
    sigma : float, complex, or numpy.ndarray, optional
    mu : float, complex, or numpy.ndarray, optional
    radius : float, optional
    current: float, optional
    turns : int, optional
    secondary : bool, optional
        whether to return the secondary field (primary if false)

    Returns
    -------
    hz : complex, or numpy.ndarray
        The secondary (primary if `secondary == False`) vertical magnetic field

    Notes
    -----
    The inputs values will be broadcasted together following normal numpy rules, and
    will support general shapes. Therefore every input, except for the `secondary` flag,
    can be arrays of the same shape.

    The analytic expression for the total magnetic field from equation 4.94 in Ward and
    Hohmann 1988:

    .. math::
        H_z = - \\frac{I}{k^2 a^3}\\left[3 - \\left(3 + 3 i k a - k^2 a^2\\right)e^{-i k a}\\right]

    Examples
    --------

    This example reproduces figure 4.7 from Ward and Hohmann

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.fdem import vertical_magnetic_field_horizontal_loop

    Define the frequency range,

    >>> frequencies = np.logspace(-1, 6, 200)
    >>> hz = vertical_magnetic_field_horizontal_loop(frequencies, sigma=1E-2, radius=50, secondary=False)

    Then plot the values

    >>> plt.loglog(frequencies, hz.real, c='C0', label='Real')
    >>> plt.loglog(frequencies, -hz.real, '--', c='C0')
    >>> plt.loglog(frequencies, hz.imag, c='C1', label='Imaginary')
    >>> plt.loglog(frequencies, -hz.imag, '--', c='C1')
    >>> plt.xlabel('frequency (Hz)')
    >>> plt.ylabel('H$_z$ (A/m)')
    >>> plt.legend()
    >>> plt.show()
    """
    w = 2*np.pi*frequencies
    k = np.sqrt(-1j * w * mu * sigma)
    a = radius

    Hz = (
        -current
        / (k ** 2 * a ** 3)
        * (3 - (3 + 3j * k * a - k ** 2 * a ** 2) * np.exp(-1j * k * a))
    )

    if secondary:
        Hzp = current / 2.0 / a
        Hz = Hz - Hzp
    return turns * Hz


def vertical_magnetic_flux_horizontal_loop(
    frequencies, sigma, mu=mu_0, radius=1.0, current=1.0, turns=1, secondary=True
):
    """Vertical magnetic flux density at the center of a horizontal loop for each frequency

    Simple function to calculate the vertical magnetic field due to a harmonic
    horizontal loop. The anlytic form is only available at the center of the loop.

    Parameters
    ----------
    frequencies : float, or numpy.ndarray
    sigma : float, complex, or numpy.ndarray
    mu : float, complex, or numpy.ndarray, optional
    radius : float, optional
    current: float, optional
    turns : int, optional
    secondary : bool, optional
        whether to return the secondary field (primary if false)

    Returns
    -------
    bz : complex, or numpy.ndarray
        The seecondary (primary if `secondary == False`) vertical magnetic flux density

    Notes
    -----
    The inputs values will be broadcasted together following normal numpy rules, and
    will support general shapes. Therefore every input, except for the `secondary` flag,
    can be arrays of the same shape.

    See Also
    --------
    vertical_magnetic_field_horizontal_loop
    """
    return mu * vertical_magnetic_field_horizontal_loop(
        frequencies, sigma, mu=mu, radius=radius, current=current, turns=turns, secondary=secondary
    )



import os
def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration("kernels", parent_package, top_path)

    # Conditionally add subpackage if intending to build compiled components
    if os.environ.get('BUILD_GEOANA_EXT', "0") != "0":
        config.add_subpackage("_extensions")

    return config
